/*
 * libsyncml - A syncml protocol implementation
 * Copyright (C) 2008  Michael Bell <michael.bell@opensync.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 */

#ifndef _SML_DATA_SYNC_TRANSPORT_H_
#define _SML_DATA_SYNC_TRANSPORT_H_

#include "data_sync.h"

/* OBEX client */

SmlBool smlDataSyncTransportObexClientInit(
		SmlDataSyncObject *dsObject,
		SmlError **error);

SmlBool smlDataSyncTransportObexClientConnect(
		SmlDataSyncObject *dsObject,
		SmlError **error);

/* HTTP client */

SmlBool smlDataSyncTransportHttpClientInit(
		SmlDataSyncObject *dsObject,
		SmlError **error);

SmlBool smlDataSyncTransportHttpClientConnect(
		SmlDataSyncObject *dsObject,
		SmlError **error);

/* HTTP server */

SmlBool smlDataSyncTransportHttpServerInit(
		SmlDataSyncObject *dsObject,
		SmlError **error);

#endif /* _SML_DATA_SYNC_TRANSPORT_H_ */

